#
# pcbe-gcp.awk - (want to) convert PCBE gerber output to RS-274X, and
#                             drill output to excellon NC drill format
#   by uratan! 2008.12.10
#
# usage:
#   gawk -f pcbe-gcp.awk -v drill=1  hout.lst w-drill.hol     > z-drill.cnc
#   gawk -f pcbe-gcp.awk -v rev=0    gout.lst w-btm-mask.GRB  > z-btm-mask.gbx
#
# give: drill ... filter as drill file if !0
#       rev   ... add reverse command output if !0
#
# tested under PCBE 0.48.2, 0.48.8
#
BEGIN {
	#
	# version
	#
	ver = "(pcbe-gcp.awk v0.1)";

	#
	# set this if clean up PCBE gerber output
	#   (1) 1 [hƂɉs
	#   (2) Ap[`I D1n ɂ G54 tB
	#   (3) D01/D02 ͒PƂŎgȂBXnnnYnnn ɂړ(ׂ)ɕt
	#   (4) D03 ͒PƂŎgp (͌̂܂܂)
	#
	to_fix_gerber_output = 1;

	#
	# initial my variables
	#
	fname = "";
	n_file = 0;

	#
	# output headers
	#
	if(drill) {
		#
		# confirm misc settings as NC drill file
		#
		print "M48";		# Program Header
		print ";;";							# comment
		print ";; digits: 3.3";				# comment
		print ";; Absolute";				# comment
		print ";; units: mm";				# comment
		print ";; Leading-Zero-Suppress";	# comment
		print ";;   " ver;					# comment
		print ";;";							# comment
		#print "M71";		# metric mode
		print "METRIC,TZ";	# metric mode, Trailing zero mode
	} else {
		#
		# confirm misc settings as RS-274X
		#
		print "G04                       *";	# comment
		print "G04 digits: 4.3           *";	# comment
		print "G04 Absolute              *";	# comment
		print "G04 units: mm             *";	# comment
		print "G04 Leading-Zero-Suppress *";	# comment
		print "G04   " ver " *";				# comment
		print "G04                       *";	# comment
		print "%FSLAX43Y43*%";	# Format statement
								# 'L'eading zeros omitted,
								# 'A'bsolute coordinates,
								# 4 integer digits, 3 fractional digits
		print "%MOMM*%";		# set unit to millimeters

		#
		#
		#
		if(to_fix_gerber_output) {
			init_fix_gerber();
		}
	}
}

{
	if(FILENAME != fname) {
		n_file++;
		fname = FILENAME;

		if((n_file == 2) && (!drill) && (rev == 1)) {
			print "%IPNEG*%";	# image polatiry - negative
		}
	}
}

{
	if(n_file == 1) {
		#
		# at first file
		#
		if(drill) {
			#
			# make drill size list from hout.lst
			#
			if(match($0, /^  T.*/)) {
				print $1 "C" $3;
			}
		} else {
			#
			# make aperture size list from gout.lst
			#
			if(match($0, "  ۃAp")) {
				sub(/.*D/, "D", $1);
				print "%AD" $1 "C," $3 "*%";
			} else if(match($0, "  pAp")) {
				sub(/.*D/, "D", $1);
				print "%AD" $1 "R," $3 "X" $3"*%";
			}
		}
	} else {
		#
		# at second file
		#
		if(drill) {
			#
			# convert PCBE output for NC drill format
			#
			if($0 == "X0Y0") {
				#
				# kill non-hole-moving
				#
				;
			} else if(match($0, "M05$")) {
				#
				# kill unknown code 'M05'
				#
				sub(/M05$/, "", $0);
				print $0;
			} else {
				sub(/M02/, "M30", $0);	# End of program
				print $0;
			}
		} else {
			if(to_fix_gerber_output) {
				#
				# modify PCBE gerber output to avoid complaint by GC-Prevue
				#
				fix_gerber($0);
			} else {
				#
				# pass orignal thru
				#
				print $0;
			}
		}
	}
}

#
# modify PCBE gerber output to avoid GC-Prevue complaint
#
function init_fix_gerber()
{
	gerber_str_remained = "";
	M_has_come = 0;

	prev_G_code = "";
	prev_X_code = "";
	prev_Y_code = "";
	prev_Dap_code = "";
	prev_D0n_code = "";
}

#
#
#
function fix_gerber(input)
{
	#
	# join strings remained in previous line
	#
	input = gerber_str_remained input;

	#
	# loop in this string
	#
	while(1) {
		#
		# abandon process once any Mnn has come.
		#
		if(M_has_come) {
			break;
		}

		#
		# search EOB
		#
		i = index(input, "*");
		if(i == 0) {
			#
			# EOB is not found, pass strings remained to next line
			#
			gerber_str_remained = input;
			break;
		}

		#
		# cut out the word (omit EOB '*')
		#
		word = substr(input, 1, i-1);
		input = substr(input, i+1);

		#
		# examine the word
		#   we will stop this process if Mnn has come
		#
		if(match(word, /M[0-9][0-9]/)) {
			M_code = substr(word, RSTART, RLENGTH);
			M_has_come = 1;
		}

		#
		#  Gnn, Xnn, Ynn, Mnn can be treated as plain static modifier
		#
		G_updated = 0;
		if(match(word, /G[0-9][0-9]*/)) {
			G_code = substr(word, RSTART, RLENGTH);
			if(prev_G_code != G_code) {
				G_updated = 1;
				prev_G_code = G_code;
			}
		} else {
			G_code = prev_G_code;
		}

		X_updated = 0;
		if(match(word, /X[0-9][0-9]*/)) {
			X_code = substr(word, RSTART, RLENGTH);
			if(prev_X_code != X_code) {
				X_updated = 1;
				prev_X_code = X_code;
			}
		} else {
			X_code = prev_X_code;
		}

		Y_updated = 0;
		if(match(word, /Y[0-9][0-9]*/)) {
			Y_code = substr(word, RSTART, RLENGTH);
			if(prev_Y_code != Y_code) {
				Y_updated = 1;
				prev_Y_code = Y_code;
			}
		} else {
			Y_code = prev_Y_code;
		}

		#
		# D1nn for selecting aperture is used single by itself
		#
		AP_changed = 0;
		if(match(word, /D[1-9][0-9]*/)) {
			Dap_code = substr(word, RSTART, RLENGTH);
			if(prev_Dap_code != Dap_code) {
				AP_changed = 1;
				prev_Dap_code = Dap_code;
			}
		} else {
			Dap_code = prev_Dap_code;
		}

		#
		# D03 for flash-aperture is also used single by itself
		#
		D03_appeared = 0;
		if(match(word, /D03/)) {
			D03_appeared = 1;
			prev_D0n_code = "D02";		# reset mode: Exposure-off
		}

		#
		# D01, D02 can be treated as plain static modifier.
		#
		if(match(word, /D0[12]/)) {
			D0n_code = substr(word, RSTART, RLENGTH);
			if(prev_D0n_code != D0n_code) {
				prev_D0n_code = D0n_code;
			}
		} else {
			D0n_code = prev_D0n_code;
		}

		#
		# make RS-274D output
		#
		if(AP_changed) {
			#
			# change aperture size, this is used as single Dnn command
			#
			printf("G54%s*\n", Dap_code);
		} else if(D03_appeared) {
			#
			# flash-aperture, this is used as single Dnn command
			#
			printf("D03*\n");
		} else {
			#
			# draw command, parameters same as previous can be omitted.
			# you can force add them like:
			#   G_updated = 1;
			#   X_updated = 1;
			#   Y_updated = 1;
			#
			if(G_updated) {
				printf("%s", G_code);
			}
			if(X_updated) {
				printf("%s", X_code);
			}
			if(Y_updated) {
				printf("%s", Y_code);
			}
			if(G_updated || X_updated || Y_updated) {
				printf("%s", D0n_code);
			}
			if(M_has_come) {
				printf("%s", M_code);
			}
			if(G_updated || X_updated || Y_updated || M_has_come) {
				printf("*\n");
			}
		}
	}
}

####################################################################
# Ql
#
# RS-274D, RS-274X
#   ΓO uK[o[tH[}bg []v
#       <http://www.nagano-it.go.jp/joho/kaisetsu/gbf_s.pdf>
#   ΓO ugK[o[tH[}bgv
#       <http://www.nagano-it.go.jp/joho/kaisetsu/ext_gerber.pdf>
#   "Gerber RS-274X Format User's Guide" Rev D March, 2001
#       <http://members.optusnet.com.au/~eseychell/rs274xrevd_e.pdf>
#
# Excellon NC drill format
#   "USER MANUALS", CNC-7 "Part Programming Codes"
#       <http://www.excellon.com/manuals/program.htm>
#
####################################################################
# PCBE gerber output parameters
#   - Pʁ@mm
#   -  4C RC_@s[TvX
#   - (Absolute coordinate)
#
# PCBE gerber output is on RS-274D, has:
#
#   G71*      specify millimeters
#   G90*      specify absolute
#   ...
#     D01*        draw line, exposure on
#     D03*        Flash aperture
#     D02*        exposure off
#     Dnn*        select an aperture
#     G01XnnYnn*  linear interpolation
#   ...
#   M00*      Program stop
#
#   - * - * - * - * -
#
# D03 ̈ɂ
#
#   ERS-274X Manual ɂ́A
#       "D03 remains in effect until a new layer is encountered." Ə
#       ̂ D03 ̓[_ł(L)ׂAƎvĂ̂낤B
#   Ȃ̂A
#   EPCBE ́A
#       D03 R[h̓[_łȂALAND łANVƎvĂB
#       (D1nn ̃Ap[`[IƓAPƂŌ)
#   EGC-Prevue A
#       D03 ̌ XY ̈ړw肳 Exposure-off move ƈB
#   Ȃ̂ŁAɂ킹A
#     EDnn ׂ͂ĒPƂŌƋK肷B
#     ED03, D1nn ׂ͂ĒPƂŏo͂B
#     ED03 ̃[h D02 ɖ߂B
#       悤ɂ悤B
#
#   +------------ ł낢eXg ---------------
#   |G04      *
#   |G04 Test *
#   |G04      *
#   |%FSLAX43Y43*%
#   |%MOMM*%
#   |%ADD22C,0.250*%
#   |%ADD33C,1.000*%
#   |G71*
#   |G90*
#   |G54D22*
#   |G01X10000Y10000D02*
#   |G01X11000Y11000*
#   |G04 --- D02 is inherited --- *
#   |X15000Y10000D01*
#   |Y15000*
#   |X10000*
#   |G04 --- D01 is also inherited --- *
#   |D03*
#   |Y10000*
#   |G04 --- D03 is NOT inherited --- *
#   |G04 --- D01 is reset, treated as D02 --- *
#   |G54D33*
#   |G01X11000Y12500D02*
#   |D03*
#   |G01X12500*
#   |G04 --- D03 is NOT inherited --- *
#   |G04 --- always treated as D02 --- *
#   |G01X14000D02*
#   |D01*
#   |D02*
#   |G04 --- D01+D02 makes single flash --- *
#   |M02*
#   +-----------------------------------------------------
#
#   Ȃ݂ GC-Prevue ł D01+D02 Ńhł邪A
#            PCBE(K[o[) ł͑łȂB
#
####################################################################
# PCBE drill output parameters
#   - Pʁ@mm
#   -  4C 3C_@s[TvX
#   - (Absolute coordinate)
#
# PCBE drill output is on (some) NC drill format, has:
#             __PCBE_expected__       __correct_meaning__
#   %          Rewind and Stop           <--
#   G90        Absolute Mode             <--
#   X0Y0       Move                    Move and Drill
#   T20        Tool selection            <--
#   XnnYnnM05  Move and Drill          M05 is undefined
#   ...
#   M02        End of program          misc extention
#
####################################################################
