#
# pcbe-mkfont.awk - make PCBEFONT.DAT from *.pcb
#   by uratan! 2008.10.31
#
# convert 5 PAGEs in total (see 'pcbe-mkfont-template.pcb')
#
#    +---------------------+
#    |                     +-+
#    |         PAGE 6        |  JIS:0x2621 -- 0x267e
#    +-+                     |      EEEEEE
#    +---------------------+
#    |                     +-+
#    |         PAGE 5        |  JIS:0x2521 -- 0x257e
#    +-+                     |      ACEGIEEE
#    +-+-------------------+-+
#    |                     +-+
#    |         PAGE 4        |  JIS:0x2421 -- 0x247e
#    +-+                     |      EEE
#    +-+-------------------+-+
#    |                     +-+
#    |         PAGE 3        |  JIS:0x2321 -- 0x237e
#    +-+                     |      OPQEE`abEE
#    +-+-------------------+-+
#    |                     +-+
#    |         PAGE 2        |  JIS:0x2221 -- 0x227e
#    +-+                     |      EEE
#    +---------------------+-+ -----------------------------
#    |                     +-+  JIS:0x2121 -- 0x217e     |
#    |         PAGE 1        |      @ABHIEEE    row:6
#    +-+-+-..                |                            |
#    o +-+-------------------+ -----------------------------
#    |                       |
#    |                       |     o: at (17.0,17.0)
#    |    coloumn:16         |        frame is 17x17 for a char of height=16
#    |<--------------------->|
#
BEGIN {
	FS = "[=,]";		# set my field separater
#	BINMODE = "w";		# expect 'no conversion on outpout', but no effect !

	#
	# constants
	#
	FRAME_STEP = 17.0;
	EXPAND_RATIO = 2.0;
	ORIGIN_X = FRAME_STEP * 1;
	ORIGIN_Y = FRAME_STEP * 1;
	N_PAGE = 6;
	MAX_N_LINES_per_CHAR = 100;		# real limit is 'n_cmds < 256'

	TAB_TOP_ADDR = 0x0080;	# vector data pointer table from jis-code
	VEC_TOP_ADDR = 0x5e00;	# font vector data

	#
	# OPTIONS
	#
	# select compression level
	#   0: no compression
	#   1: compress only start position by seeking line connection
	#      (reduce vector data about 30%)
	#   2: compress all X,Y coordinates individually
	#      (reduce vector data about 40%)
	#
	OPT_COMPRESS_LEVEL = 2;
	#
	# permit single DOT by 0-length-line (valid only OPT_COMPRESS_LEVEL=2)
	#
	OPT_PERMIT_0_LEN_LINE = 1;
	#
	# print misc info (! this may interfere xxd.exe !)
	#
	DEBUG = 0;

	#
	# initial global variables
	#
	Out_Addr = 0x0000;		# variable Out_Addr is shared with xxd_dump()

	#
	# output header
	#
	xxd_dump(0x50);		# 'P'
	xxd_dump(0x43);		# 'C'
	xxd_dump(0x42);		# 'B'
	xxd_dump(0x45);		# 'E'
	xxd_dump(0x20);		# ' '
	xxd_dump(0x46);		# 'F'
	xxd_dump(0x4F);		# 'O'
	xxd_dump(0x4E);		# 'N'
	xxd_dump(0x54);		# 'T'
	xxd_dump(0x20);		# ' '
	xxd_dump(0x44);		# 'D'
	xxd_dump(0x41);		# 'A'
	xxd_dump(0x54);		# 'T'
	xxd_dump(0x41);		# 'A'

	while(Out_Addr < TAB_TOP_ADDR) {
		xxd_dump(0x00);
	}
}

#
# load all lines on layer 0:⏕
#
/^LINE/ {
	# $1      $2      $3 $4      $5 $6     $7 $8      $9 $10 $11 $12
	# LINE:XS 305.000 YS 441.000 XE 35.000 YE 441.000 LE 1   SQ  28
	#
	if($10 == 0) {
		xs = $2;
		ys = $4;
		js = coor_to_jis(xs, ys);
		xe = $6;
		ye = $8;
		je = coor_to_jis(xe, ye);

		if((js == je) && (js != 0)) {
			n_line = int(vect[js "N_LINE"]);
			if(n_line < MAX_N_LINES_per_CHAR) {
				xs = int((xs - Local_Origin_X) * EXPAND_RATIO);
				ys = int((ys - Local_Origin_Y) * EXPAND_RATIO);
				xe = int((xe - Local_Origin_X) * EXPAND_RATIO);
				ye = int((ye - Local_Origin_Y) * EXPAND_RATIO);
				if((xs < 32) && (ys < 32) && (xe < 32) && (ye < 32)) {
					vect[js "L" n_line "XS"] = xs;
					vect[js "L" n_line "YS"] = ys;
					vect[js "L" n_line "XE"] = xe;
					vect[js "L" n_line "YE"] = ye;
					vect[js "N_LINE"] = n_line + 1;
				}
			}
		}
	}
	next;
}

#
# dump all data
#
END {
	#
	# output jis-code to data pointer table
	#   loop for each page
	#
	vec_addr = VEC_TOP_ADDR;
	for(page=0; page<N_PAGE; page++) {
		#
		# loop for in page
		#
		for(jis_L=0x21; jis_L<=0x7e; jis_L++) {
			jis = 0x2100 + page * 0x100 + jis_L;

			n_line = vect[jis "N_LINE"];
			if(n_line != 0) {
				#
				# generate vector data
				#
				if(OPT_COMPRESS_LEVEL == 0) {
					plain_convert_vectors(jis, n_line);
				} else {
					sort_by_connection(jis, n_line);
				}
			} else {
				vect[jis "N_BYTES"] = 0;
			}

			n_bytes = vect[jis "N_BYTES"];
			if(n_bytes == 0) {
				#
				# this code has NO font data
				#
				xxd_dump(0x00);
				xxd_dump(0x00);
				xxd_dump(0x00);
			} else {
				#
				# this code has font data
				#
				xxd_dump(int(vec_addr / 0x00001) % 0x100);
				xxd_dump(int(vec_addr / 0x00100) % 0x100);
				xxd_dump(int(vec_addr / 0x10000) % 0x100);

				#
				# advance vector data pointer
				#
				vec_addr += 1 + n_bytes;
			}
		}
	}

	#
	# fill unused table area
	#
	while(Out_Addr < VEC_TOP_ADDR) {
		xxd_dump(0x00);
	}

	#
	# flush vector data
	#   loop for each page
	#
	for(page=0; page<N_PAGE; page++) {
		#
		# loop for in page
		#
		for(jis_L=0x21; jis_L<=0x7e; jis_L++) {
			jis = 0x2100 + page * 0x100 + jis_L;

			n_bytes = vect[jis "N_BYTES"];
			if(n_bytes != 0) {
				#
				# output number of commands
				#
				xxd_dump(vect[jis "N_CMDS"]);

				#
				# output vector data
				#
				for(i=0; i<n_bytes; i++) {
					xxd_dump(vect[jis "RAW" i]);
				}
			}
		}
	}
}

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

#
# use xxd.exe (in VIM package) to convert this dump to binary
#
#   if your gawk works correct about BINMODE = "w",
#   you can make binary output directly.
#    (try to examine 'printf("%c", 0x0a);')
#
function xxd_dump(data)
{
	if((Out_Addr % 0x10) == 0x00) {
		printf("%07x:", Out_Addr);
	}

	printf(" %02x", data);

	if((Out_Addr % 0x10) == 0x0f) {
		printf("\n");
	}

	Out_Addr++;
}

#
# map given coordinates to JIS code
#
function coor_to_jis(x, y)
{
	cj_nx = int((x - ORIGIN_X) / FRAME_STEP);
	cj_ny = int((y - ORIGIN_Y) / FRAME_STEP);

	if((cj_nx < 0) || (16 <= cj_nx))
		return 0;
	if((cj_ny < 0) || ((6 * N_PAGE) <= cj_ny))
		return 0;

	Local_Origin_X = ORIGIN_X + cj_nx * FRAME_STEP;
	Local_Origin_Y = ORIGIN_Y + cj_ny * FRAME_STEP;

	cj_ny_div = int(cj_ny / 6);
	cj_ny_mod = cj_ny % 6;
	cj_jis = 0x20 + cj_ny_mod * 0x10 + cj_nx;
	if((cj_jis == 0x20) || (cj_jis == 0x7f))
		return 0;

	cj_jis += 0x2100 + cj_ny_div * 0x0100;
	return cj_jis;
}

#
# convert lines to vector format with no modification
#
function plain_convert_vectors(jis, n_line)
{
	pc_n_cmds = 0;
	pc_n_bytes = 0;

	for(pc_l=0; pc_l<n_line; pc_l++) {
		#
		# output 'move' command
		#
		vect[jis "RAW" pc_n_bytes++] = 0x60 + vect[jis "L" pc_l "XS"];
		vect[jis "RAW" pc_n_bytes++] =        vect[jis "L" pc_l "YS"];
		pc_n_cmds++;

		#
		# output 'draw' command
		#
		vect[jis "RAW" pc_n_bytes++] = 0xe0 + vect[jis "L" pc_l "XE"];
		vect[jis "RAW" pc_n_bytes++] =        vect[jis "L" pc_l "YE"];
		pc_n_cmds++;
	}

	vect[jis "N_CMDS"] = pc_n_cmds;
	vect[jis "N_BYTES"] = pc_n_bytes;
}

#
# sort lines by connection
#
function sort_by_connection(jis, n_line)
{
	#
	# make a single string which has all lines
	#   fwd means the lines are described 'start-to-end'
	#   rvs means the lines are described 'end-to-start'
	#
	sc_fwd_list = "";
	sc_rvs_list = "";
	for(sc_l=0; sc_l<n_line; sc_l++) {
		sc_xs = sprintf("%02d", vect[jis "L" sc_l "XS"]);
		sc_ys = sprintf("%02d", vect[jis "L" sc_l "YS"]);
		sc_xe = sprintf("%02d", vect[jis "L" sc_l "XE"]);
		sc_ye = sprintf("%02d", vect[jis "L" sc_l "YE"]);
		sc_fwd_list = sc_fwd_list " " sc_xs "," sc_ys "-" sc_xe "," sc_ye;
		sc_rvs_list = sc_rvs_list " " sc_xe "," sc_ye "-" sc_xs "," sc_ys;
	}

	#
	# each element in the list is fixed length
	#
	N_CHARS_per_LINE = length(" xs,ys-xe,ye");
	N_CHARS_per_COOR = N_CHARS_per_LINE / 2;
if(DEBUG) {
 printf("\n");
 printf("==== jis:%04x (n_line:%d) ====\n", jis, n_line);
 printf(" fwd_list:%s\n", sc_fwd_list);
 printf(" rvs_list:%s\n", sc_rvs_list);
}

	#
	# start to sort
	#
	sc_n_cmds = 0;
	sc_n_bytes = 0;
	sc_prev_x = 0;
	sc_prev_y = 0;

	#
	# loop until all lines are examined
	#
	while(sc_fwd_list != "") {
		#
		# take out the head line of the list.
		# then remove the line from the lists.
		#
		sc_a_line_fwd = substr(sc_fwd_list, 1, N_CHARS_per_LINE);
		sc_a_line_rvs = substr(sc_rvs_list, 1, N_CHARS_per_LINE);
		sub(sc_a_line_fwd, "", sc_fwd_list);
		sub(sc_a_line_rvs, "", sc_rvs_list);

		#
		# initial list of connected lines
		#
		sc_conn_list = sc_a_line_fwd;

		#
		# cut out start and end coordinate of the connected-lines.
		#   both strings are start with " "(SPACE)
		#
		sc_start_coor = substr(sc_a_line_fwd, 1, N_CHARS_per_COOR);
		sc_end_coor   = substr(sc_a_line_rvs, 1, N_CHARS_per_COOR);
if(DEBUG) {
 printf("0:a_line_fwd: %s\n", sc_a_line_fwd);
 printf("0:a_line_rvs: %s\n", sc_a_line_rvs);
 printf(" fwd_list: %s\n", sc_fwd_list);
 printf(" rvs_list: %s\n", sc_rvs_list);
 printf(" ...conn_list: %s\n", sc_conn_list);
 printf("\n");
}

		#
		# scan lines remained in the list
		#
		while(sc_fwd_list != "") {
if(DEBUG)printf("(%s)-...-(%s)\n", sc_start_coor, sc_end_coor);
			#
			# search a line connected
			#
			if((sc_i = index(sc_fwd_list, sc_start_coor)) != 0) {
				sc_found = 1;
			}
			else if((sc_i = index(sc_rvs_list, sc_start_coor)) != 0) {
				sc_found = 2;
			}
			else if((sc_i = index(sc_fwd_list, sc_end_coor)) != 0) {
				sc_found = 3;
			}
			else if((sc_i = index(sc_rvs_list, sc_end_coor)) != 0) {
				sc_found = 4;
			} else {
				break;		# not found
			}

			#
			# cut out the line matched.
			# then remove the line from the lists.
			#
			sc_i = int((sc_i-1) / N_CHARS_per_LINE) * N_CHARS_per_LINE + 1;
			sc_a_line_fwd = substr(sc_fwd_list, sc_i, N_CHARS_per_LINE);
			sc_a_line_rvs = substr(sc_rvs_list, sc_i, N_CHARS_per_LINE);
			sub(sc_a_line_fwd, "", sc_fwd_list);
			sub(sc_a_line_rvs, "", sc_rvs_list);

			#
			# connect the line and update start/end coordinate
			#
			if(sc_found == 1) {
				sc_conn_list = sc_a_line_rvs sc_conn_list;
				sc_start_coor = substr(sc_a_line_rvs, 1, N_CHARS_per_COOR);
			} else if(sc_found == 2) {
				sc_conn_list = sc_a_line_fwd sc_conn_list;
				sc_start_coor = substr(sc_a_line_fwd, 1, N_CHARS_per_COOR);
			} else if(sc_found == 3) {
				sc_conn_list = sc_conn_list sc_a_line_fwd;
				sc_end_coor = substr(sc_a_line_rvs, 1, N_CHARS_per_COOR);
			} else {	# if(sc_found == 4)
				sc_conn_list = sc_conn_list sc_a_line_rvs;
				sc_end_coor = substr(sc_a_line_fwd, 1, N_CHARS_per_COOR);
			}
if(DEBUG) {
 printf("n:sc_found: %d\n", sc_found);
 printf("n:a_line_fwd: %s\n", sc_a_line_fwd);
 printf("n:a_line_rvs: %s\n", sc_a_line_rvs);
 printf(" fwd_list: %s\n", sc_fwd_list);
 printf(" rvs_list: %s\n", sc_rvs_list);
 printf(" ...conn_list: %s\n\n", sc_conn_list);
}
		}

		#
		# flush connected lines as vector format
		#
		sc_nl = length(sc_conn_list) / N_CHARS_per_LINE;
		sc_p = 1;
		for(sc_l=0; sc_l<sc_nl; sc_l++) {
			sc_xs = int(substr(sc_conn_list, sc_p+1, 2));
			sc_ys = int(substr(sc_conn_list, sc_p+4, 2));
			sc_xe = int(substr(sc_conn_list, sc_p+7, 2));
			sc_ye = int(substr(sc_conn_list, sc_p+10, 2));
			sc_p += N_CHARS_per_LINE;

			if(OPT_COMPRESS_LEVEL == 1) {
				sc_sub_flush_with_weak_compression(jis);
			} else {
				sc_sub_flush_with_full_compression(jis, OPT_PERMIT_0_LEN_LINE);
			}
		}
	}

	vect[jis "N_CMDS"] = sc_n_cmds;
	vect[jis "N_BYTES"] = sc_n_bytes;
}

#
# flush a line as vector format with weak compression
#  ! all variables sc_xxx are shared with sort_by_connection() !
#
function sc_sub_flush_with_weak_compression(jis)
{
	#
	# output 'move' command
	#   this is suppressed when the start-position is at last end-position
	#
	if((sc_xs != sc_prev_x) || (sc_ys != sc_prev_y)) {
		vect[jis "RAW" sc_n_bytes++] = 0x60 + sc_xs;
		vect[jis "RAW" sc_n_bytes++] =        sc_ys;
		sc_n_cmds++;
	}

	#
	# output 'draw' command
	#
	vect[jis "RAW" sc_n_bytes++] = 0xe0 + sc_xe;
	vect[jis "RAW" sc_n_bytes++] =        sc_ye;
	sc_n_cmds++;

	#
	# memory end-position
	#
	sc_prev_x = sc_xe;
	sc_prev_y = sc_ye;
}

#
# flush a line as vector format with full compression
#  ! all variables sc_xxx are shared with sort_by_connection() !
#
function sc_sub_flush_with_full_compression(jis, permit_0_length_line)
{
	#
	# output 'move' command
	#
	if((sc_xs == sc_prev_x) && (sc_ys == sc_prev_y)) {
		;
	}
	else if(sc_xs == sc_prev_x) {
		vect[jis "RAW" sc_n_bytes++] = 0x40 + sc_ys;
		sc_n_cmds++;
	}
	else if(sc_ys == sc_prev_y) {
		vect[jis "RAW" sc_n_bytes++] = 0x20 + sc_xs;
		sc_n_cmds++;
	}
	else {
		vect[jis "RAW" sc_n_bytes++] = 0x60 + sc_xs;
		vect[jis "RAW" sc_n_bytes++] =        sc_ys;
		sc_n_cmds++;
	}
	sc_prev_x = sc_xs;
	sc_prev_y = sc_ys;

	#
	# output 'draw' command
	#
	if((sc_xe == sc_prev_x) && (sc_ye == sc_prev_y)) {
		if(permit_0_length_line) {
			vect[jis "RAW" sc_n_bytes++] = 0xc0 + sc_ye;
			sc_n_cmds++;
		} else {
			;
		}
	}
	else if(sc_xe == sc_prev_x) {
		vect[jis "RAW" sc_n_bytes++] = 0xc0 + sc_ye;
		sc_n_cmds++;
	}
	else if(sc_ye == sc_prev_y) {
		vect[jis "RAW" sc_n_bytes++] = 0xa0 + sc_xe;
		sc_n_cmds++;
	}
	else {
		vect[jis "RAW" sc_n_bytes++] = 0xe0 + sc_xe;
		vect[jis "RAW" sc_n_bytes++] =        sc_ye;
		sc_n_cmds++;
	}
	sc_prev_x = sc_xe;
	sc_prev_y = sc_ye;
}

##############################################################################
#
# /*
#  * PCBE font file format
#  *   by uratan! 2008.10.30
#  */
#
# --------
# ABSTRUCT
# --------
#
#     0x0000-0x007F : header
#     0x0080-0x5DFF : JIS code to vector data address table
#     0x5E00-       : vector data
#
#
# ------
# HEADER
# ------
#
#     string "PCBE FONT DATA" is here.
#
#     *EXAMPLE*
#         0000 : 50 43 42 45 20 46 4F 4E 54 20 44 41 54 41 00 00
#         0010 : 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
#         0020 : 00 ....
#           .
#           .
#         0070 : 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
#
#
# ------------------------
# JIS TO VECTOR DATA TABLE
# ------------------------
#
#     1) JIS code from 0x2121 to 0x747e is mapped on 0x0080 to 0x5d07.
#     2) address is described by 3bytes. (little-endian)
#     3) address 0x000000 specifies UNDEFINED. (treated as BLANK/SPACE)
#     4) table address and JIS code conversion is done like:
#         tab_addr = 0x0080 + ((JIScodeH-0x21) * 94 + (JIScodeL-0x21)) * 3
#         JIScodeH = ((tab_addr - 0x0080) / 3) / 94 + 0x21
#         JIScodeL = ((tab_addr - 0x0080) / 3) % 94 + 0x21
#
#     *EXAMPLE*
#         0080 : 00-00-00 00-5E-00 09-5E-00 18-5E-00 23-5E-00 2C-
#                ~~2121~~ ~~2122~~ ~~2123~~ ~~2124~~ ~~2125~~ ~~
#         0090 : 5E-00 39-5E-00 4A-5E-00 5D-5E-00 79-5E-00 8A-5E-
#               2126~~ ~~2127~~ ~~2128~~ ~~2129~~ ~~212a~~ ~~212b
#         00A0 : 00 90-5E-00 00-00-00 9F-5E-00 00-00-00 A4-5E-00
#                ~~ ~~212c~~ ~~212d~~ ~~212e~~ ~~212f~~ ~~2130~~
#
#
# ------------------
# VECTOR DATA FORMAT
# ------------------
#
#     1) number of commands at first byte.
#     2) command are followed, it is defined like below.
#
#         CMD: pen move commands like
#             +---+---+---+---++---+---+---+---+
#             | p | y | x | c || c | c | c | c |
#             +---+---+---+---++---+---+---+---+
#                 p: 0 : pen up
#                    1 : pen down
#                 y: 0 : no Y coordinate
#                    1 : has Y coordinate
#                 x: 0 : no X coordinate
#                    1 : has X coordinate
#                 c:   : coodinate (0 - 31)
#
#         DAT: Y coordinate if needed
#             +---+---+---+---++---+---+---+---+
#             | 0 | 0 | 0 | c || c | c | c | c |
#             +---+---+---+---++---+---+---+---+
#                 c:   : coodinate (0 - 31)
#
#     3) where, font data coordinate is:
#               ^
#             Y | - - - - - * (31,31)
#               |
#               |           |
#               |
#               |           |
#               +------------> X
#             (0,0)
#
#     *EXAMPLE*
#         '^' (0x213F)
#             02, 62-02 FC-1C
#                     2 commands, init at 0,0
#                     move to X:2,Y:2
#                     draw to X:28,Y:28
#
#         'm' (0x214E)
#             04, 32 AC DE B2
#                     4 commands, init at 0,0
#                     move to X:18
#                     draw to X:12
#                     draw to Y:30
#                     draw to X:18
#
#         'I' (0x212A)
#             0A, 6E-0C DC 46 F0-04 EE-02 EC-04 EE-06 C2 70-04 AC
#                     10 commands, init at 0,0
#                     move to X:14,Y:12
#                     draw to Y:28
#                     move to Y:6
#                     draw to X:16,Y:4
#                     draw to X:14,Y:2
#                     draw to X:12,Y:4
#                     draw to X:14,Y:6
#                     draw to Y:2
#                     move to X:16,Y:4
#                     draw to X:12
#
#         '' (0x2174)
#             08, 62-08 BA 72-02 F6-1C 7C-16 A4 6C-1C E8 02
#                     8 commands, init at 0,0
#                     move to X:2,Y:8
#                     draw to X:26
#                     move to X:18,Y:2
#                     draw to X:22,Y:28
#                     move to X:28,Y:22
#                     draw to X:4
#                     move to X:12,Y:28
#                     draw to X:8,Y:2
#
#         'O' (0x2330)
#             09, 64-08 EA-02 B4 FA-08 D6 F4-1C AA E4-16 C8
#                     9 commands, init at 0,0
#                     move to X:4,Y:8
#                     draw to X:10,Y:2
#                     draw to X:20
#                     draw to X:26,Y:8
#                     draw to Y:22
#                     draw to X:20,Y:28
#                     draw to X:10
#                     draw to X:4,Y:22
#                     draw to Y:8
#
#
# -----
# NOTEs
# -----
#
#     1) single-byte ASCII chars are mapped to JIS code.
#     2) KATAKANA font is included in distribution.
#     3) 2dot(?) spacing is automatically inserted between chars.
#     4) default installed font list is below: (ver.0.48.2)
#             (XX means undefined in JIS code)
#
#     DEFAULT INSTALLED FONT                             JIS CODE CHARS
#        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#        |  |A|B|C|D|E|F|G|H|I|J|K|  |M|  |     |@|A|B|C|D|E|F|G|H|I|J|K|L|M|N|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |O|  |Q|  |  |  |  |  |  |  |  |  |[|  |  |^|  |O|P|Q|R|S|T|U|V|W|X|Y|Z|[|\|]|^|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |  |`|  |b|  |  |  |f|  |h|i|j|  |  |m|n|  |_|`|a|b|c|d|e|f|g|h|i|j|k|l|m|n|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |o|p|  |  |  |  |u|v|  |  |  |  |{|||  |  |  |o|p|q|r|s|t|u|v|w|x|y|z|{|||}|~|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |  ||  |||  |  |  |  |  |  |  |  |  |  ||  |||||||||||||||||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     ||  |  ||||||  |  |  |  |  |  |  |     ||||||||||||||||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#
#        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#        |  |  |  |  |  |  |  |  |  |  |  |  |  |  |XX|     |||||||||||||||XX|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|  |  |  |  |  |  |  |XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|||||||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |  |  |XX|XX|XX|XX|XX|XX|XX|XX|  |  |  |  |  |  |  |||XX|XX|XX|XX|XX|XX|XX|XX|||||||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |  |XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|  |  |  |  |  ||XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|||||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |  |  |  |  |  |  |  |  |  |  |  |XX|XX|XX|XX|XX|  ||||||||||||XX|XX|XX|XX|XX|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |XX|XX|  |  |  |  |  |  |  |  |XX|XX|XX|XX|  |     |XX|XX|||||||||XX|XX|XX|XX||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#
#        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#        |XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|     |XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |O|P|Q|R|S|T|U|V|W|X|XX|XX|XX|XX|XX|XX|  |O|P|Q|R|S|T|U|V|W|X|XX|XX|XX|XX|XX|XX|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |XX|`|a|b|c|d|e|f|g|h|i|j|k|l|m|n|  |XX|`|a|b|c|d|e|f|g|h|i|j|k|l|m|n|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |o|p|q|r|s|t|u|v|w|x|y|XX|XX|XX|XX|XX|  |o|p|q|r|s|t|u|v|w|x|y|XX|XX|XX|XX|XX|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |XX||||||||||||||||  |XX||||||||||||||||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     ||||||||||||XX|XX|XX|XX|     ||||||||||||XX|XX|XX|XX|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#
#        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#        |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |     ||||||||||||||||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |||||||||||||||||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |||||||||||||||||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |||||||||||||||||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |||||||||||||||||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |  |  |  |  |XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|     |||||XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|XX|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#
#        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#        |@|A|B|C|D|E|F|G|H|I|J|K|L|M|N|     |@|A|B|C|D|E|F|G|H|I|J|K|L|M|N|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |O|P|Q|R|S|T|U|V|W|X|Y|Z|[|\|]|^|  |O|P|Q|R|S|T|U|V|W|X|Y|Z|[|\|]|^|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |_|`|a|b|c|d|e|f|g|h|i|j|k|l|m|n|  |_|`|a|b|c|d|e|f|g|h|i|j|k|l|m|n|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |o|p|q|r|s|t|u|v|w|x|y|z|{|||}|~|  |o|p|q|r|s|t|u|v|w|x|y|z|{|||}|~|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |||||||||||||||  ||  |||||||||||||||||
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#     |  |  |||  |  |  |XX|XX|XX|XX|XX|XX|XX|XX|     ||||||||XX|XX|XX|XX|XX|XX|XX|XX|
#     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#
##############################################################################
